import pandas as pd
import numpy as np


def save_track_layers(
    tracks_directory,
    points_layers,
    track_file_extension=".points",
):
    print(f"Saving tracks to: {tracks_directory}")
    tracks_directory.mkdir(parents=True, exist_ok=True)

    for points_layer in points_layers:
        save_single_track(
            points_layer.data,
            points_layer.name,
            tracks_directory,
            track_file_extension=track_file_extension,
        )


def save_single_track(
    points,
    name,
    output_directory,
    track_file_extension=".points",
):
    output_filename = output_directory / (name + track_file_extension)
    points = pd.DataFrame(points)
    points.to_hdf(output_filename, key="df", mode="w")


def export_splines(tracks_directory, splines, spline_names, resolution):
    print(f"Exporting tracks to: {tracks_directory}")
    tracks_directory.mkdir(parents=True, exist_ok=True)

    for spline, name in zip(splines, spline_names):
        export_single_spline(spline, name, tracks_directory, resolution)


def export_single_spline(
    spline,
    name,
    output_directory,
    resolution,
    spline_file_extension=".npy",
):
    output_filename = output_directory / (name + spline_file_extension)
    np.save(str(output_filename), spline * resolution)



import napari
import numpy as np
from pathlib import Path
from napari.qt.threading import thread_worker
from qtpy import QtCore

from qtpy.QtWidgets import (
    QLabel,
    QFileDialog,
    QGridLayout,
    QGroupBox,
    QWidget,
)

from bg_atlasapi import BrainGlobeAtlas

from brainreg_segment.paths import Paths

from brainreg_segment.regions.IO import (
    save_label_layers,
    export_label_layers,
)

from brainreg_segment.tracks.IO import save_track_layers, export_splines

from brainreg_segment.atlas.utils import (
    get_available_atlases,
    structure_from_viewer,
)
from brainreg_segment.layout.utils import display_warning

# LAYOUT HELPERS ################################################################################

# from brainreg_segment.layout.utils import (
#     disable_napari_key_bindings,
#     disable_napari_btns,
#     overwrite_napari_roll,
# )
from brainreg_segment.layout.gui_constants import (
    WINDOW_HEIGHT,
    WINDOW_WIDTH,
    COLUMN_WIDTH,
    SEGM_METHODS_PANEL_ALIGN,
    LOADING_PANEL_ALIGN,
    BOUNDARIES_STRING,
    TRACK_FILE_EXT,
    DISPLAY_REGION_INFO,
)

from brainreg_segment.layout.gui_elements import (
    add_button,
    add_combobox,
)

# SEGMENTATION  ################################################################################
from brainreg_segment.segmentation_panels.regions import RegionSeg
from brainreg_segment.segmentation_panels.tracks import TrackSeg


class SegmentationWidget(QWidget):
    def __init__(
        self,
        viewer: napari.viewer.Viewer,
        boundaries_string=BOUNDARIES_STRING,
    ):
        super(SegmentationWidget, self).__init__()

        # general variables
        self.viewer = viewer

        # Disable / overwrite napari viewer functions
        # that either do not make sense or should be avoided by the user
        # removed for now, to make sure plugin
        # disable_napari_btns(self.viewer)
        # disable_napari_key_bindings()
        # overwrite_napari_roll(self.viewer)

        # Main layers
        self.base_layer = []  # Contains registered brain / reference brain
        self.atlas_layer = []  # Contains annotations / region information

        # Other data
        self.hemispheres_data = []

        # Track variables
        self.track_layers = []

        # Region variables
        self.label_layers = []

        # Atlas variables
        self.current_atlas_name = ""
        self.atlas = None

        self.boundaries_string = boundaries_string
        self.directory = ""
        # Set up segmentation methods
        self.region_seg = RegionSeg(self)
        self.track_seg = TrackSeg(self)

        # Generate main layout
        self.setup_main_layout()

        if DISPLAY_REGION_INFO:

            @self.viewer.mouse_move_callbacks.append
            def display_region_info(v, event):
                """
                Show brain region info on mouse over in status bar on the right
                """
                assert self.viewer == v
                if v.dims.ndisplay == 2:
                    if len(v.layers) and self.atlas_layer and self.atlas:
                        _, _, _, region_info = structure_from_viewer(
                            self.viewer.status, self.atlas_layer, self.atlas
                        )
                        self.viewer.help = region_info
                else:
                    self.viewer.help = ""

    def setup_main_layout(self):
        """
        Construct main layout of widget
        """
        self.layout = QGridLayout()
        self.layout.setContentsMargins(10, 10, 10, 10)
        self.layout.setAlignment(QtCore.Qt.AlignTop)
        self.layout.setSpacing(4)

        # 3 Steps:
        # - Loading panel
        # - Segmentation methods panel
        # -> Individual segmentation methods (which are invisible at first)
        # - Saving panel

        self.add_loading_panel(1)
        self.add_segmentation_methods_panel(1)
        self.track_seg.add_track_panel(2)  # Track segmentation subpanel
        self.region_seg.add_region_panel(3)  # Region segmentation subpanel
        self.add_saving_panel(4)

        # Take care of status label
        self.status_label = QLabel()
        self.status_label.setText("Ready")
        self.layout.addWidget(self.status_label, 5, 0)

        self.setLayout(self.layout)

    # PANELS ###############################################################

    def add_segmentation_methods_panel(self, row, column=1):
        """
        Segmentation methods chooser panel:
            Toggle visibility of segmentation
            methods
        """
        self.toggle_methods_panel = QGroupBox("Segmentation")
        self.toggle_methods_layout = QGridLayout()
        self.toggle_methods_layout.setContentsMargins(10, 10, 10, 10)
        self.toggle_methods_layout.setSpacing(5)
        self.toggle_methods_layout.setAlignment(QtCore.Qt.AlignBottom)

        self.show_trackseg_button = add_button(
            "Track tracing",
            self.toggle_methods_layout,
            self.track_seg.toggle_track_panel,
            0,
            1,
            minimum_width=COLUMN_WIDTH,
            alignment=SEGM_METHODS_PANEL_ALIGN,
        )
        self.show_trackseg_button.setEnabled(False)

        self.show_regionseg_button = add_button(
            "Region segmentation",
            self.toggle_methods_layout,
            self.region_seg.toggle_region_panel,
            1,
            1,
            minimum_width=COLUMN_WIDTH,
            alignment=SEGM_METHODS_PANEL_ALIGN,
        )
        self.show_regionseg_button.setEnabled(False)

        self.toggle_methods_layout.setColumnMinimumWidth(1, COLUMN_WIDTH)
        self.toggle_methods_panel.setLayout(self.toggle_methods_layout)
        self.toggle_methods_panel.setVisible(True)

        self.layout.addWidget(self.toggle_methods_panel, row, column, 1, 1)

    def add_loading_panel(self, row, column=0):
        """
        Loading panel:
            - Load project (sample space)
            - Load project (atlas space)
            - Atlas chooser
        """
        self.load_data_panel = QGroupBox("Load data")
        self.load_data_layout = QGridLayout()
        self.load_data_layout.setSpacing(15)
        self.load_data_layout.setContentsMargins(10, 10, 10, 10)
        self.load_data_layout.setAlignment(QtCore.Qt.AlignBottom)

        self.load_button = add_button(
            "Load project (sample space)",
            self.load_data_layout,
            self.load_brainreg_directory_sample,
            0,
            0,
            visibility=False,
            minimum_width=COLUMN_WIDTH,
            alignment=LOADING_PANEL_ALIGN,
        )

        self.load_button_standard = add_button(
            "Load project (atlas space)",
            self.load_data_layout,
            self.load_brainreg_directory_standard,
            1,
            0,
            visibility=False,
            minimum_width=COLUMN_WIDTH,
            alignment=LOADING_PANEL_ALIGN,
        )

        self.add_atlas_menu(self.load_data_layout)

        self.load_data_layout.setColumnMinimumWidth(0, COLUMN_WIDTH)
        self.load_data_panel.setLayout(self.load_data_layout)
        self.load_data_panel.setVisible(True)

        self.layout.addWidget(self.load_data_panel, row, column, 1, 1)

        #  buttons made visible after adding to main widget, preventing them
        # from briefly appearing in a separate window
        self.load_button.setVisible(True)
        self.load_button_standard.setVisible(True)

    def add_saving_panel(self, row):
        """
        Saving/Export panel
        """
        self.save_data_panel = QGroupBox()
        self.save_data_layout = QGridLayout()

        self.export_button = add_button(
            "To brainrender",
            self.save_data_layout,
            self.export_to_brainrender,
            0,
            0,
            visibility=False,
        )
        self.save_button = add_button(
            "Save", self.save_data_layout, self.save, 0, 1, visibility=False
        )

        self.save_data_layout.setColumnMinimumWidth(1, COLUMN_WIDTH)
        self.save_data_panel.setLayout(self.save_data_layout)
        self.layout.addWidget(self.save_data_panel, row, 0, 1, 2)

        self.save_data_panel.setVisible(False)

    # ATLAS INTERACTION ####################################################

    def add_atlas_menu(self, layout):
        list_of_atlasses = ["Load atlas"]
        available_atlases = get_available_atlases()
        for atlas in available_atlases.keys():
            atlas_desc = f"{atlas} v{available_atlases[atlas]}"
            list_of_atlasses.append(atlas_desc)
            atlas_menu, _ = add_combobox(
                layout,
                None,
                list_of_atlasses,
                2,
                0,
                label_stack=True,
                callback=self.initialise_atlas,
                width=COLUMN_WIDTH,
            )

        self.atlas_menu = atlas_menu

    def initialise_atlas(self):
        atlas_string = self.atlas_menu.currentText()
        atlas_name = atlas_string.split(" ")[0].strip()
        if atlas_name != self.current_atlas_name:
            status = self.remove_layers()
            if not status:  # Something prevented deletion
                self.reset_atlas_menu()
                return
        else:
            print(f"{atlas_string} already selected for segmentation.")
            self.reset_atlas_menu()
            return

        # Get / set output directory
        self.set_output_directory()
        if not self.directory:
            self.reset_atlas_menu()
            return

        self.current_atlas_name = atlas_name
        # Instantiate atlas layers
        self.load_atlas()

        self.directory = self.directory / atlas_name
        self.paths = Paths(self.directory, atlas_space=True)

        self.status_label.setText("Ready")
        # Set window title
        # self.viewer.title = f"Atlas: {self.current_atlas_name}"
        self.initialise_segmentation_interface()
        # Check / load previous regions and tracks
        self.region_seg.check_saved_region()
        self.track_seg.check_saved_track()
        self.reset_atlas_menu()

    def set_output_directory(self):
        self.status_label.setText("Loading...")
        options = QFileDialog.Options()
        options |= QFileDialog.DontUseNativeDialog
        self.directory = QFileDialog.getExistingDirectory(
            self,
            "Select output directory",
            options=options,
        )
        if self.directory != "":
            self.directory = Path(self.directory)

    def load_atlas(self):
        atlas = BrainGlobeAtlas(self.current_atlas_name)
        self.atlas = atlas
        self.base_layer = self.viewer.add_image(
            self.atlas.reference,
            name="Reference",
        )
        self.atlas_layer = self.viewer.add_labels(
            self.atlas.annotation,
            name=self.atlas.atlas_name,
            blending="additive",
            opacity=0.3,
            visible=False,
        )
        self.standard_space = True

    def reset_atlas_menu(self):
        # Reset menu for atlas - show initial description
        self.atlas_menu.blockSignals(True)
        self.atlas_menu.setCurrentIndex(0)
        self.atlas_menu.blockSignals(False)

    # BRAINREG INTERACTION #################################################

    def load_brainreg_directory_sample(self):
        self.get_brainreg_directory(standard_space=False)

    def load_brainreg_directory_standard(self):
        self.get_brainreg_directory(standard_space=True)

    def get_brainreg_directory(self, standard_space):
        """
        Shows file dialog to choose output directory
        and sets global directory info
        """
        if standard_space:
            self.plugin = "brainreg-standard"
            self.standard_space = True
        else:
            self.plugin = "brainglobe-io"
            self.standard_space = False

        self.status_label.setText("Loading...")
        options = QFileDialog.Options()
        options |= QFileDialog.DontUseNativeDialog
        brainreg_directory = QFileDialog.getExistingDirectory(
            self,
            "Select brainreg directory",
            options=options,
        )

        if not brainreg_directory:
            return

        if self.directory != brainreg_directory:
            status = self.remove_layers()
            if not status:
                return  # Something prevented deletion
            self.directory = Path(brainreg_directory)
        else:
            print(f"{str(brainreg_directory)} already loaded.")
            return

        # Otherwise, proceed loading brainreg dir
        self.load_brainreg_directory()

    def load_brainreg_directory(self):
        """
        Opens brainreg folder in napari.
        Calls initialise_loaded_data to set up layers / info.
        Then checks for previously loaded data.

        """
        try:
            self.viewer.open(str(self.directory), plugin=self.plugin)
            self.paths = Paths(
                self.directory,
                standard_space=self.standard_space,
            )
            self.initialise_loaded_data()
        except ValueError:
            print(
                f"The directory ({self.directory}) does not appear to be "
                f"a brainreg directory, please try again."
            )
            return

        # Check / load previous regions and tracks
        self.region_seg.check_saved_region()
        self.track_seg.check_saved_track()

    def initialise_loaded_data(self):
        """
        Set up brainreg layers in napari / fill with new data and info

        """
        try:
            self.viewer.layers.remove(self.boundaries_string)
        except ValueError:
            pass

        self.base_layer = self.viewer.layers["Registered image"]
        self.metadata = self.base_layer.metadata
        self.atlas = self.metadata["atlas_class"]
        self.atlas_layer = self.viewer.layers[self.metadata["atlas"]]
        if self.standard_space:
            self.hemispheres_data = self.atlas.hemispheres
        else:
            self.hemispheres_data = self.viewer.layers["Hemispheres"].data

        self.initialise_segmentation_interface()

        # Set window title
        # self.viewer.title = (
        #     f"Brainreg: {self.metadata['atlas']} ({self.plugin})"
        # )
        self.status_label.setText("Ready")

    # MORE LAYOUT COMPONENTS ###########################################

    def initialise_segmentation_interface(self):
        self.reset_variables()
        self.initialise_image_view()
        self.save_data_panel.setVisible(True)
        self.save_button.setVisible(True)
        self.export_button.setVisible(self.standard_space)
        self.show_regionseg_button.setEnabled(True)
        self.show_trackseg_button.setEnabled(True)
        self.status_label.setText("Ready")

    def initialise_image_view(self):
        self.set_z_position()

    def set_z_position(self):
        midpoint = int(round(len(self.base_layer.data) / 2))
        self.viewer.dims.set_point(0, midpoint)

    def reset_variables(self):
        """
        Reset atlas scale dependent variables
        - point_size (Track segmentation)
        - spline_size (Track segmentation)
        - brush_size (Region segmentation)
        """
        self.mean_voxel_size = int(
            np.sum(self.atlas.resolution) / len(self.atlas.resolution)
        )
        self.track_seg.point_size = (
            self.track_seg.point_size_default / self.mean_voxel_size
        )
        self.track_seg.spline_size = (
            self.track_seg.spline_size_default / self.mean_voxel_size
        )
        self.region_seg.brush_size = (
            self.region_seg.brush_size_default / self.mean_voxel_size
        )
        return

    def remove_layers(self):
        """
        TODO: This needs work. Runs into an error currently
        when switching from a annotated project to another one
        """
        if len(self.viewer.layers) != 0:
            # Check with user if that is really what is wanted
            if self.track_layers or self.label_layers:
                choice = display_warning(
                    self,
                    "About to remove layers",
                    "All layers are about to be deleted. Proceed?",
                )
                if not choice:
                    print('Preventing deletion because user chose "Cancel"')
                    return False

            # Remove old layers
            for layer in list(self.viewer.layers):
                try:
                    self.viewer.layers.remove(layer)
                except IndexError:  # no idea why this happens
                    pass

        # There seems to be a napari bug trying to access previously used slider
        # values. Trying to circument for now
        self.viewer.window.qt_viewer.dims._last_used = None

        self.track_layers = []
        self.label_layers = []
        return True

    def save(self):
        if self.label_layers or self.track_layers:
            choice = display_warning(
                self,
                "About to save files",
                "Existing files will be will be deleted. Proceed?",
            )
            if choice:
                print("Saving")
                worker = save_all(
                    self.paths.regions_directory,
                    self.paths.tracks_directory,
                    self.label_layers,
                    self.track_layers,
                    track_file_extension=TRACK_FILE_EXT,
                )
                worker.start()
            else:
                print('Not saving because user chose "Cancel" \n')

    def export_to_brainrender(self):
        choice = display_warning(
            self,
            "About to export files",
            "Existing files will be will be deleted. Proceed?",
        )
        if choice:
            print("Exporting")
            worker = export_all(
                self.paths.regions_directory,
                self.paths.tracks_directory,
                self.label_layers,
                self.track_seg.splines,
                self.track_seg.spline_names,
                self.atlas.resolution[0],
            )
            worker.start()
        else:
            print('Not exporting because user chose "Cancel" \n')


@thread_worker
def export_all(
    regions_directory,
    tracks_directory,
    label_layers,
    splines,
    spline_names,
    resolution,
):
    if label_layers:
        export_label_layers(regions_directory, label_layers, resolution)

    if splines:
        export_splines(tracks_directory, splines, spline_names, resolution)
    print("Finished!\n")


@thread_worker
def save_all(
    regions_directory,
    tracks_directory,
    label_layers,
    points_layers,
    track_file_extension=".points",
):

    if label_layers:
        save_label_layers(regions_directory, label_layers)

    if points_layers:
        save_track_layers(
            tracks_directory,
            points_layers,
            track_file_extension=track_file_extension,
        )
    print("Finished!\n")


def main():
    print("Loading segmentation GUI.\n ")
    with napari.gui_qt():
        viewer = napari.Viewer()  # title="Segmentation GUI")
        viewer.window.resize(WINDOW_WIDTH, WINDOW_HEIGHT)
        widget = SegmentationWidget(viewer)
        viewer.window.add_dock_widget(widget, name="General", area="right")


if __name__ == "__main__":
    main()



import tifffile
from filecmp import cmp

from pathlib import Path
from brainreg_segment.regions import IO as region_IO


regions_dir = Path.cwd() / "tests" / "data" / "regions"
VOXEL_SIZE = 100


def test_export_regions_to_file(tmpdir):
    image = tifffile.imread(regions_dir / "region.tiff")
    filename = tmpdir / "region.obj"
    region_IO.export_regions_to_file(image, filename, VOXEL_SIZE)

    cmp(regions_dir / "region.obj", tmpdir / "region.obj")



import shutil

# import tifffile

import numpy as np
import pandas as pd

from pathlib import Path
from filecmp import cmp
from brainreg_segment.segment import SegmentationWidget

brainreg_dir = Path.cwd() / "tests" / "data" / "brainreg_output"

ATLAS_NAME = "example_mouse_100um"


def test_load_sample_space(make_test_viewer):
    viewer = make_test_viewer()
    widget = SegmentationWidget(viewer)
    viewer.window.add_dock_widget(widget, name="General", area="right")
    widget.standard_space = False
    widget.plugin = "brainreg"
    widget.directory = brainreg_dir
    widget.load_brainreg_directory()
    check_loaded_layers(widget, 6)


def test_load_standard_space(make_test_viewer):
    viewer = make_test_viewer()
    widget = SegmentationWidget(viewer)
    viewer.window.add_dock_widget(widget, name="General", area="right")
    widget.standard_space = True
    widget.plugin = "brainreg_standard"
    widget.directory = brainreg_dir
    widget.load_brainreg_directory()
    check_loaded_layers(widget, 4)


def test_load_atlas(tmpdir, make_test_viewer):
    viewer = make_test_viewer()
    widget = SegmentationWidget(viewer)
    viewer.window.add_dock_widget(widget, name="General", area="right")
    widget.directory = tmpdir
    widget.current_atlas_name = ATLAS_NAME
    widget.load_atlas()
    assert len(widget.viewer.layers) == 2
    assert widget.base_layer.name == "Reference"
    assert widget.atlas.atlas_name == ATLAS_NAME
    assert widget.atlas_layer.name == widget.atlas.atlas_name


def test_general(make_test_viewer):
    viewer = make_test_viewer()
    widget = SegmentationWidget(viewer)
    viewer.window.add_dock_widget(widget, name="General", area="right")
    widget.standard_space = True
    widget.plugin = "brainreg_standard"
    widget.directory = brainreg_dir
    widget.load_brainreg_directory()
    assert widget.mean_voxel_size == 50
    check_defaults(widget)
    check_paths(widget)


def test_tracks(tmpdir, make_test_viewer, rtol=1e-10):
    tmp_input_dir = tmpdir / "brainreg_output"
    test_tracks_dir = (
        tmp_input_dir / "manual_segmentation" / "standard_space" / "tracks"
    )
    validate_tracks_dir = (
        brainreg_dir / "manual_segmentation" / "standard_space" / "tracks"
    )
    shutil.copytree(brainreg_dir, tmp_input_dir)
    viewer = make_test_viewer()
    widget = SegmentationWidget(viewer)
    viewer.window.add_dock_widget(widget, name="General", area="right")
    widget.standard_space = True
    widget.plugin = "brainreg_standard"
    widget.directory = Path(tmp_input_dir)
    widget.load_brainreg_directory()

    assert len(widget.viewer.layers) == 4
    assert len(widget.track_layers) == 1
    widget.track_seg.add_track()
    assert len(widget.viewer.layers) == 5
    assert len(widget.track_layers) == 2
    assert widget.track_layers[0].name == "test_track"
    assert widget.track_layers[1].name == "track_1"
    assert len(widget.track_layers[0].data) == 6

    # analysis
    widget.track_seg.run_track_analysis()
    regions_validate = pd.read_csv(validate_tracks_dir / "test_track.csv")
    regions_test = pd.read_csv(test_tracks_dir / "test_track.csv")
    pd.testing.assert_frame_equal(regions_validate, regions_test)

    # saving
    widget.save()
    points_validate = pd.read_hdf(validate_tracks_dir / "test_track.points")
    points_test = pd.read_hdf(test_tracks_dir / "test_track.points")
    np.testing.assert_allclose(points_validate, points_test, rtol=rtol)

    # export
    widget.export_to_brainrender()
    spline_validate = pd.read_hdf(validate_tracks_dir / "test_track.h5")
    spline_test = pd.read_hdf(test_tracks_dir / "test_track.h5")
    pd.testing.assert_frame_equal(spline_test, spline_validate)

    # surface points
    widget.track_seg.add_surface_points()
    assert len(widget.track_layers[0].data) == 7


def test_regions(tmpdir, make_test_viewer, rtol=1e-10):
    tmp_input_dir = tmpdir / "brainreg_output"
    test_regions_dir = (
        tmp_input_dir / "manual_segmentation" / "standard_space" / "regions"
    )
    validate_regions_dir = (
        brainreg_dir / "manual_segmentation" / "standard_space" / "regions"
    )
    shutil.copytree(brainreg_dir, tmp_input_dir)
    viewer = make_test_viewer()
    widget = SegmentationWidget(viewer)
    viewer.window.add_dock_widget(widget, name="General", area="right")
    widget.standard_space = True
    widget.plugin = "brainreg_standard"
    widget.directory = Path(tmp_input_dir)
    widget.load_brainreg_directory()

    assert len(widget.viewer.layers) == 4
    assert len(widget.label_layers) == 1
    widget.region_seg.add_region()
    assert len(widget.viewer.layers) == 5
    assert len(widget.label_layers) == 2
    assert widget.label_layers[0].name == "test_region"
    assert widget.label_layers[1].name == "region_1"

    # analysis
    widget.region_seg.run_region_analysis()
    region_csv_validate = pd.read_csv(validate_regions_dir / "test_region.csv")
    region_csv_test = pd.read_csv(test_regions_dir / "test_region.csv")
    pd.testing.assert_frame_equal(region_csv_test, region_csv_validate)

    summary_csv_validate = pd.read_csv(validate_regions_dir / "summary.csv")
    summary_csv_test = pd.read_csv(test_regions_dir / "summary.csv")
    pd.testing.assert_frame_equal(summary_csv_test, summary_csv_validate)

    # saving
    widget.save()

    # image_validate = tifffile.imread(validate_regions_dir / "test_region.tiff")
    # image_test = tifffile.imread(test_regions_dir / "test_region.tiff")
    # np.testing.assert_allclose(image_validate, image_test, rtol=rtol)

    # export
    widget.export_to_brainrender()
    cmp(
        validate_regions_dir / "test_region.obj",
        test_regions_dir / "test_region.obj",
    )


def check_loaded_layers(widget, num_layers):
    assert len(widget.viewer.layers) == num_layers
    assert widget.base_layer.name == "Registered image"
    assert widget.atlas.atlas_name == "allen_mouse_50um"
    assert widget.metadata["orientation"] == "psl"
    assert widget.atlas_layer.name == widget.atlas.atlas_name


def check_defaults(widget):
    assert widget.track_seg.point_size == int(100 / widget.mean_voxel_size)
    assert widget.track_seg.spline_size == int(50 / widget.mean_voxel_size)
    assert widget.track_seg.track_file_extension == ".points"
    assert widget.region_seg.image_file_extension == ".tiff"
    assert widget.region_seg.num_colors == 10
    assert widget.region_seg.brush_size == int(250 / widget.mean_voxel_size)
    assert widget.track_seg.spline_points_default == 1000
    assert widget.track_seg.spline_smoothing_default == 0.1
    assert widget.track_seg.fit_degree_default == 3
    assert widget.track_seg.summarise_track_default is True
    assert widget.region_seg.calculate_volumes_default is True
    assert widget.region_seg.summarise_volumes_default is True
    assert widget.boundaries_string == "Boundaries"


def check_paths(widget):
    assert widget.paths.main_directory == brainreg_dir / "manual_segmentation"

    assert (
        widget.paths.segmentation_directory
        == brainreg_dir / "manual_segmentation" / "standard_space"
    )

    assert (
        widget.paths.regions_directory
        == brainreg_dir / "manual_segmentation" / "standard_space" / "regions"
    )

    assert (
        widget.paths.region_summary_csv
        == brainreg_dir
        / "manual_segmentation"
        / "standard_space"
        / "regions"
        / "summary.csv"
    )
    assert (
        widget.paths.tracks_directory
        == brainreg_dir / "manual_segmentation" / "standard_space" / "tracks"
    )



import pandas as pd
from pathlib import Path


def add_new_track_layer(viewer, track_layers, point_size):
    num = len(track_layers)
    new_track_layers = viewer.add_points(
        None,
        ndim=3,
        n_dimensional=True,
        size=point_size,
        name=f"track_{num}",
    )
    new_track_layers.mode = "ADD"
    track_layers.append(new_track_layers)


def add_existing_track_layers(viewer, track_file, point_size):
    points = pd.read_hdf(track_file)
    new_points_layer = viewer.add_points(
        points,
        n_dimensional=True,
        size=point_size,
        name=Path(track_file).stem,
    )
    new_points_layer.mode = "ADD"
    return new_points_layer



# GUI ELEMENTS
# from napari.resources import build_icons # Contains .SVGPATH to all icons for napari

from qtpy.QtWidgets import (
    QDoubleSpinBox,
    QPushButton,
    QCheckBox,
    QLabel,
    QSpinBox,
    QComboBox,
)


def add_combobox(
    layout,
    label,
    items,
    row,
    column=0,
    label_stack=False,
    callback=None,
    width=150,
):
    if label_stack:
        combobox_row = row + 1
        combobox_column = column
    else:
        combobox_row = row
        combobox_column = column + 1
    combobox = QComboBox()
    combobox.addItems(items)
    if callback:
        combobox.currentIndexChanged.connect(callback)
    combobox.setMaximumWidth = width

    if label is not None:
        combobox_label = QLabel(label)
        combobox_label.setMaximumWidth = width
        layout.addWidget(combobox_label, row, column)
    else:
        combobox_label = None

    layout.addWidget(combobox, combobox_row, combobox_column)
    return combobox, combobox_label


def add_button(
    label,
    layout,
    connected_function,
    row,
    column,
    visibility=True,
    minimum_width=0,
    alignment="center",
):
    button = QPushButton(label)
    if alignment == "center":
        pass
    elif alignment == "left":
        button.setStyleSheet("QPushButton { text-align: left; }")
    elif alignment == "right":
        button.setStyleSheet("QPushButton { text-align: right; }")

    button.setVisible(visibility)
    button.setMinimumWidth(minimum_width)
    layout.addWidget(button, row, column)
    button.clicked.connect(connected_function)
    return button


# def add_radiobutton(
#     label,
#     layout,
#     connected_function,
#     row,
#     column,
#     visibility=True,
#     minimum_width=0,
#     alignment="center",
# ):
#     button = QRadioButton(label)
#     if alignment == "center":
#         pass
#     elif alignment == "left":
#         button.setStyleSheet(
#             "QRadioButton { text-align: left; padding: 0; spacing: 30px;}"
#         )
#     elif alignment == "right":
#         button.setStyleSheet(
#             "QRadioButton { text-align: right; padding: 0; spacing: 30px;}"
#         )

#     # Too change indicator button ... needs to dynamically retrieve icon from Napari.
#     # Icons are saved as .svg files under napari.resources SVGPATH
#     # "QRadioButton::indicator"
#     # "{"
#     # "width:16px;"
#     # "height:16px;"
#     # "}"
#     # "QRadioButton::indicator::unchecked"
#     # "{"
#     # "image: url(build_icons.SVGPATH/visibility_off.svg);"
#     # "}"
#     # "QRadioButton::indicator::checked"
#     # "{"
#     # "image: url(/opt/miniconda3/envs/analysis/lib/python3.6/site-packages/napari/resources/icons/visibility.svg);"
#     # "}"
#     # )

#     button.setVisible(visibility)
#     button.setMinimumWidth(minimum_width)
#     layout.addWidget(button, row, column)
#     button.clicked.connect(connected_function)
#     return button


def add_checkbox(layout, default, label, row, column=0):
    box = QCheckBox()
    box.setChecked(default)
    layout.addWidget(QLabel(label), row, column)
    layout.addWidget(box, row, column + 1)
    return box


def add_float_box(
    layout, default, minimum, maximum, label, step, row, column=0
):
    box = QDoubleSpinBox()
    box.setMinimum(minimum)
    box.setMaximum(maximum)
    box.setValue(default)
    box.setSingleStep(step)
    layout.addWidget(QLabel(label), row, column)
    layout.addWidget(box, row, column + 1)
    return box


def add_int_box(layout, default, minimum, maximum, label, row, column=0):
    box = QSpinBox()
    box.setMinimum(minimum)
    box.setMaximum(maximum)
    # Not always set if not after min & max
    box.setValue(default)
    layout.addWidget(QLabel(label), row, column)
    layout.addWidget(box, row, column + 1)
    return box



# COPIED FROM NAPARI - how to import?

import warnings
import pytest

from typing import List

from qtpy.QtWidgets import QApplication
from napari import Viewer


def pytest_addoption(parser):
    """Add napari specific command line options.

    --show-viewer
        Show viewers during tests, they are hidden by default. Showing viewers
        decreases test speed by around 20%.

    --perfmon-only
        Run only perfmon test.

    Notes
    -----
    Due to the placement of this conftest.py file, you must specifically name
    the napari folder such as "pytest napari --show-viewer"

    For --perfmon-only must also enable perfmon with env var:
    NAPARI_PERFMON=1 pytest napari --perfmon-only
    """
    parser.addoption(
        "--show-viewer",
        action="store_true",
        default=False,
        help="don't show viewer during tests",
    )

    parser.addoption(
        "--perfmon-only",
        action="store_true",
        default=False,
        help="run only perfmon tests",
    )


@pytest.fixture
def qtbot(qtbot):
    """A modified qtbot fixture that makes sure no widgets have been leaked."""
    initial = QApplication.topLevelWidgets()
    yield qtbot
    QApplication.processEvents()
    leaks = set(QApplication.topLevelWidgets()).difference(initial)
    # still not sure how to clean up some of the remaining vispy
    # vispy.app.backends._qt.CanvasBackendDesktop widgets...
    # if any([n.__class__.__name__ != "CanvasBackendDesktop" for n in leaks]):
    #    raise AssertionError(f"Widgets leaked!: {leaks}")
    if leaks:
        warnings.warn(f"Widgets leaked!: {leaks}")


@pytest.fixture(scope="function")
def make_test_viewer(qtbot, request):
    viewers: List[Viewer] = []

    def actual_factory(*model_args, **model_kwargs):
        model_kwargs["show"] = model_kwargs.pop(
            "show", request.config.getoption("--show-viewer")
        )
        viewer = Viewer(*model_args, **model_kwargs)
        viewers.append(viewer)
        return viewer

    yield actual_factory

    for viewer in viewers:
        viewer.close()



import numpy as np


def generate_random_points_3d_layer(shape=(10, 3)):
    data = np.random.rand(*shape) * 100.0
    return [(data, {"name": "random 3d"}, "points")]


def generate_random_points_2d_layer(shape=(10, 2)):
    data = np.random.rand(*shape) * 100.0
    return [(data, {"name": "random 2d"}, "points")]



import numpy as np
from brainreg_segment.atlas import utils as atlas_utils

from bg_atlasapi import BrainGlobeAtlas

atlas_name = "allen_mouse_50um"


def test_get_available_atlases():
    atlases = atlas_utils.get_available_atlases()

    # arbitrary selection of atlases
    assert float(atlases["allen_mouse_10um"]) >= 0.3
    assert float(atlases["allen_mouse_25um"]) >= 0.3
    assert float(atlases["allen_mouse_50um"]) >= 0.3
    assert float(atlases["mpin_zfish_1um"]) >= 0.4


def test_lateralise_atlas_image():
    atlas = BrainGlobeAtlas(atlas_name)

    mask = np.random.random(atlas.annotation.shape) > 0.7
    masked_annotations = mask * atlas.annotation
    annotations_left, annotations_right = atlas_utils.lateralise_atlas_image(
        masked_annotations,
        atlas.hemispheres,
        left_hemisphere_value=atlas.left_hemisphere_value,
        right_hemisphere_value=atlas.right_hemisphere_value,
    )

    # not the best way to test this is functioning properly
    total_vals_in = np.prod(mask.shape)
    total_vals_out = len(annotations_left) + len(annotations_right)

    assert total_vals_in == total_vals_out



import pandas as pd
import numpy as np
from brainreg_segment.tracks.fit import spline_fit


def track_analysis(
    viewer,
    atlas,
    tracks_directory,
    track_layers,
    napari_spline_size,
    spline_points=100,
    fit_degree=3,
    spline_smoothing=0.05,
    summarise_track=True,
):
    tracks_directory.mkdir(parents=True, exist_ok=True)

    print(
        f"Fitting splines with {spline_points} segments, of degree "
        f"'{fit_degree}' to the points"
    )
    splines = []
    spline_names = []

    for track_layer in track_layers:
        if len(track_layer.data) != 0:
            spline = run_track_analysis(
                track_layer.data,
                track_layer.name,
                tracks_directory,
                atlas,
                summarise_track=summarise_track,
                spline_smoothing=spline_smoothing,
                spline_points=spline_points,
                fit_degree=fit_degree,
            )

            splines.append(spline)

            viewer.add_points(
                spline,
                size=napari_spline_size,
                edge_color="cyan",
                face_color="cyan",
                blending="additive",
                opacity=0.7,
                name=track_layer.name + "_fit",
            )

            spline_names.append(track_layer.name)
    return splines, spline_names


def run_track_analysis(
    points,
    track_name,
    tracks_directory,
    atlas,
    spline_smoothing=0.05,
    spline_points=100,
    fit_degree=3,
    summarise_track=True,
):
    """
    For each set of points, run a spline fit, and (if required) determine which
     atlas region each part of the spline fit is within
    :param points: 3D numpy array of points
    :param track_name: Name of the set of points (used for saving results)
    :param tracks_directory: Where to save the results to
    :param atlas: brainglobe atlas class
    :param spline_smoothing: spline fit smoothing factor
    :param spline_points: How many points used to define the resulting
    interpolated path
    :param fit_degree: spline fit degree
    :param summarise_track: If True, save a csv with the atlas region for
    all parts of the spline fit
    :return np.array: spline fit
    """
    # Duplicate points causes fit ValueError
    # 2 stage process to ensure ordering
    _, indices = np.unique(points, return_index=True, axis=0)
    points = points[np.sort(indices)]

    spline = spline_fit(
        points,
        smoothing=spline_smoothing,
        k=fit_degree,
        n_points=spline_points,
    )
    if summarise_track:
        summary_csv_file = tracks_directory / (track_name + ".csv")
        analyse_track_anatomy(atlas, spline, summary_csv_file)

    return spline


def analyse_track_anatomy(atlas, spline, file_path):
    """
    For a given spline, find the atlas region that each
    "segment" is in, and save to csv.

    :param atlas: brainglobe atlas class
    :param spline: numpy array defining the spline interpolation
    :param file_path: path to save the results to
    :param bool verbose: Whether to print the progress
    """
    spline_regions = []
    for p in spline.tolist():
        try:
            spline_regions.append(
                atlas.structures[atlas.structure_from_coords(p)]
            )
        except KeyError:
            spline_regions.append(None)

    df = pd.DataFrame(
        columns=["Position", "Region ID", "Region acronym", "Region name"]
    )
    for idx, spline_region in enumerate(spline_regions):
        if spline_region is None:
            df = df.append(
                {
                    "Position": idx,
                    "Region ID": "Not found in brain",
                    "Region acronym": "Not found in brain",
                    "Region name": "Not found in brain",
                },
                ignore_index=True,
            )
        else:
            df = df.append(
                {
                    "Position": idx,
                    "Region ID": spline_region["id"],
                    "Region acronym": spline_region["acronym"],
                    "Region name": spline_region["name"],
                },
                ignore_index=True,
            )
    df.to_csv(file_path, index=False)



# TrackSeg
from glob import glob
import numpy as np

from qtpy.QtWidgets import (
    QGridLayout,
    QGroupBox,
)

from brainreg_segment.layout.gui_elements import (
    add_button,
    add_checkbox,
    add_float_box,
    add_int_box,
)
from brainreg_segment.tracks.layers import (
    add_new_track_layer,
    add_existing_track_layers,
)
from brainreg_segment.image.utils import create_KDTree_from_image

from brainreg_segment.tracks.analysis import track_analysis

from brainreg_segment.layout.utils import display_warning
from brainreg_segment.layout.gui_constants import (
    COLUMN_WIDTH,
    SEGM_METHODS_PANEL_ALIGN,
    POINT_SIZE,
    SPLINE_SIZE,
    TRACK_FILE_EXT,
    SPLINE_POINTS_DEFAULT,
    SPLINE_SMOOTHING_DEFAULT,
    FIT_DEGREE_DEFAULT,
    SUMMARISE_TRACK_DEFAULT,
)


class TrackSeg(QGroupBox):
    """
    Track segmentation method panel

    """

    def __init__(
        self,
        parent,
        point_size=POINT_SIZE,
        spline_size=SPLINE_SIZE,
        track_file_extension=TRACK_FILE_EXT,
        spline_points_default=SPLINE_POINTS_DEFAULT,
        spline_smoothing_default=SPLINE_SMOOTHING_DEFAULT,
        fit_degree_default=FIT_DEGREE_DEFAULT,
        summarise_track_default=SUMMARISE_TRACK_DEFAULT,
    ):

        super(TrackSeg, self).__init__()
        self.parent = parent
        self.tree = None

        self.summarise_track_default = summarise_track_default

        # Point / Spline fitting settings
        self.point_size_default = POINT_SIZE  # Keep track of default
        self.point_size = point_size  # Initialise
        self.spline_points_default = spline_points_default
        self.spline_size_default = SPLINE_SIZE  # Keep track of default
        self.spline_size = spline_size  # Initialise
        self.spline_smoothing_default = spline_smoothing_default
        self.fit_degree_default = fit_degree_default

        # File formats
        self.track_file_extension = track_file_extension

        # Initialise spline and spline names
        self.splines = None
        self.spline_names = None

    def add_track_panel(self, row):
        self.track_panel = QGroupBox("Track tracing")
        track_layout = QGridLayout()

        add_button(
            "Add surface points",
            track_layout,
            self.add_surface_points,
            5,
            1,
        )

        add_button(
            "Add track",
            track_layout,
            self.add_track,
            6,
            0,
        )

        add_button(
            "Trace tracks",
            track_layout,
            self.run_track_analysis,
            6,
            1,
        )

        self.summarise_track_checkbox = add_checkbox(
            track_layout,
            self.summarise_track_default,
            "Summarise",
            0,
        )

        self.fit_degree = add_int_box(
            track_layout,
            self.fit_degree_default,
            1,
            5,
            "Fit degree",
            1,
        )

        self.spline_smoothing = add_float_box(
            track_layout,
            self.spline_smoothing_default,
            0,
            1,
            "Spline smoothing",
            0.1,
            2,
        )

        self.spline_points = add_int_box(
            track_layout,
            self.spline_points_default,
            1,
            10000,
            "Spline points",
            3,
        )

        track_layout.setColumnMinimumWidth(1, COLUMN_WIDTH)
        self.track_panel.setLayout(track_layout)
        self.parent.layout.addWidget(self.track_panel, row, 0, 1, 2)
        self.track_panel.setVisible(False)

    def toggle_track_panel(self):
        # TODO: Change color scheme directly when theme is switched
        # TODO: "text-align" property should follow constant SEGM_METHODS_PANEL_ALIGN
        if self.track_panel.isVisible():
            self.track_panel.setVisible(False)
            if self.parent.viewer.theme == "dark":
                self.parent.show_trackseg_button.setStyleSheet(
                    f"QPushButton {{ background-color: #414851; text-align:{SEGM_METHODS_PANEL_ALIGN};}}"
                    f"QPushButton:pressed {{ background-color: #414851; text-align:{SEGM_METHODS_PANEL_ALIGN};}}"
                )
            else:
                self.parent.show_trackseg_button.setStyleSheet(
                    f"QPushButton {{ background-color: #d6d0ce; text-align:{SEGM_METHODS_PANEL_ALIGN};}}"
                    f"QPushButton:pressed {{ background-color: #d6d0ce; text-align:{SEGM_METHODS_PANEL_ALIGN};}}"
                )

        else:
            self.track_panel.setVisible(True)
            if self.parent.viewer.theme == "dark":
                self.parent.show_trackseg_button.setStyleSheet(
                    f"QPushButton {{ background-color: #7e868f; text-align:{SEGM_METHODS_PANEL_ALIGN};}}"
                    f"QPushButton:pressed {{ background-color: #7e868f; text-align:{SEGM_METHODS_PANEL_ALIGN};}}"
                )
            else:
                self.parent.show_trackseg_button.setStyleSheet(
                    f"QPushButton {{ background-color: #fdf194; text-align:{SEGM_METHODS_PANEL_ALIGN};}}"
                    f"QPushButton:pressed {{ background-color: #fdf194; text-align:{SEGM_METHODS_PANEL_ALIGN};}}"
                )

    def check_saved_track(self):
        track_files = glob(
            str(self.parent.paths.tracks_directory)
            + "/*"
            + self.track_file_extension
        )
        if self.parent.paths.tracks_directory.exists() and track_files != []:
            for track_file in track_files:
                self.parent.track_layers.append(
                    add_existing_track_layers(
                        self.parent.viewer,
                        track_file,
                        self.point_size,
                    )
                )

    def add_track(self):
        print("Adding a new track\n")
        self.splines = None
        self.spline_names = None
        self.track_panel.setVisible(True)  # Should be visible by default!
        add_new_track_layer(
            self.parent.viewer,
            self.parent.track_layers,
            self.point_size,
        )

    def add_surface_points(self):
        if self.parent.track_layers:
            print("Adding surface points (this may take a while)")
            if self.tree is None:
                self.create_brain_surface_tree()

            for track_layer in self.parent.track_layers:
                try:
                    _, index = self.tree.query(track_layer.data[0])
                except IndexError:
                    print(
                        f"{track_layer.name} does not appear to hold any data"
                    )
                    continue
                surface_point = self.tree.data[index]
                track_layer.data = np.vstack((surface_point, track_layer.data))
            print("Finished!\n")
        else:
            print("No tracks found.")

    def create_brain_surface_tree(self):
        self.tree = create_KDTree_from_image(self.parent.atlas_layer.data)

    def run_track_analysis(self):
        if self.parent.track_layers:
            choice = display_warning(
                self.parent,
                "About to analyse tracks",
                "Existing files will be will be deleted. Proceed?",
            )
            if choice:
                print("Running track analysis")
                self.splines, self.spline_names = track_analysis(
                    self.parent.viewer,
                    self.parent.atlas,
                    self.parent.paths.tracks_directory,
                    self.parent.track_layers,
                    self.spline_size,
                    spline_points=self.spline_points.value(),
                    fit_degree=self.fit_degree.value(),
                    spline_smoothing=self.spline_smoothing.value(),
                    summarise_track=self.summarise_track_checkbox.isChecked(),
                )
                print("Finished!\n")
            else:
                print("Preventing analysis as user chose 'Cancel'")
        else:
            print("No tracks found.")



import re
from bg_atlasapi.list_atlases import utils, descriptors


def lateralise_atlas_image(
    masked_atlas_annotations,
    hemispheres,
    left_hemisphere_value=1,
    right_hemisphere_value=2,
):
    """
    :param masked_atlas_annotations: Masked image of atlas annotations
    :param hemispheres: Hemispheres image
    :param left_hemisphere_value: Value encoded in hemispheres image
    :param right_hemisphere_value: Value encoded in hemispheres image
    :return: Tuple (left, right) of numpy arrays of values within
    each hemisphere
    """
    annotation_left = masked_atlas_annotations[
        hemispheres == left_hemisphere_value
    ]
    annotation_right = masked_atlas_annotations[
        hemispheres == right_hemisphere_value
    ]
    return annotation_left, annotation_right


def get_available_atlases():
    """
    Get the available brainglobe atlases
    :return: Dict of available atlases (["name":version])
    """
    available_atlases = utils.conf_from_url(
        descriptors.remote_url_base.format("last_versions.conf")
    )
    available_atlases = dict(available_atlases["atlases"])
    return available_atlases


def structure_from_viewer(status, atlas_layer, atlas):
    """
    Get brain region info from mouse position in napari viewer.

    Extract nx3 coordinate pair from napari window status string.
    Return brainglobe (BG) structure number, name, hemisphere, and a
    "pretty" string that can be displayed for example in the status bar.

    Parameter
    ---------
    status        : str, Napari viewer status (napari.viewer.Viewer.status)
    atlas_layer   : Napari viewer layer
                    Layer, which contains the annotation / region
                    information for every structure in the (registered)
                    atlas
    atlas         : Brainglobe atlas (bg_atlasapi.bg_atlas.BrainGlobeAtlas)

    Returns
    -------
    If any error is raised, (None,None,None,"") is returned
    structure_no  : int
                    BG Structure number
                    Returns none if not found
    structure     : str
                    Structure name
                    Returns none if not found
    hemisphere    : str
                    Hemisphere name
                    Returns none if not found
    region_info   : str
                    A string containing info about structure
                    and hemisphere
                    Returns empty string if not found
    """

    # Using a regex, extract list of coordinates from status string
    assert hasattr(atlas_layer, "data"), "Atlas layer appears to be empty"
    assert (
        atlas_layer.data.ndim == 3
    ), f'Atlas layer data does not have the right dim ("{atlas_layer.data.ndim}")'

    try:
        coords = re.findall(r"\[\d{1,5}\s+\d{1,5}\s+\d{1,5}\]", status)[0][
            1:-1
        ]
        coords_list = coords.split()
        map_object = map(int, coords_list)
        coord_list = tuple(map_object)
    except (IndexError, ValueError):
        # No coordinates could be extracted from status
        return None, None, None, ""

    # Extract structure number
    try:
        structure_no = atlas_layer.data[coord_list]
    except IndexError:
        return None, None, None, ""

    if structure_no in [0]:  # 0 is "Null" region
        return None, None, None, ""

    # Extract structure information
    try:
        structure = atlas.structures[structure_no]["name"]
    except KeyError:
        return None, None, None, ""

    # ... and make string pretty
    region_info = []
    for struct in structure.split(","):
        region_info.append(struct.strip().capitalize())
    hemisphere = atlas.hemisphere_from_coords(
        coord_list, as_string=True
    ).capitalize()
    region_info.append(hemisphere)
    region_info = " | ".join(region_info)

    return structure_no, structure, hemisphere, region_info



import csv
import math
import os
import pathlib
from typing import List, Tuple

from magicgui import magic_factory
import napari
import numpy as np


# For the axis This assumes the following (x,y,z) are each of the following
#  (anteroposterior, mediolateral, vectical)
def get_angles_and_distances_between_consecutive_points(
    point_layer: "napari.layers.Points",
) -> Tuple[List[float]]:
    """Returns the distance and angles  for each pair of consecutive points.

    Receives a points layer and for each pair of consecutive points it
    generates a vector, the anterioposterior and mediolateral angles of those
    vectors are returned together with the magnitude of the vectors.

    Args:
        point_layer: A napari point layer with at least two points.
    Returns:
        Three lists, one contains the anterioposterior angles, the other
        contains the mediolateral angles and the last one contains the distance
        between each points. All the lists are of length = point_layer - 1
    """
    points_array = point_layer.data

    n_points = points_array.shape[0]
    n_coords = points_array.shape[1]

    if n_coords == 2:
        points_array = np.hstack([points_array, np.zeros((n_points, 1))])

    ap_angles = []
    mp_angles = []
    distances = []
    for i in range(1, n_points):
        pointa = points_array[i - 1, :]
        pointb = points_array[i, :]
        segment_vector = pointa - pointb
        # projecting the vector in the sagital plane is just removing
        #  the "y/mediolateral" coordinate
        segment_vector_zx = segment_vector[[2, 0]]
        # for the angle, z is used as the first direction/axis and x as the other
        ap_angle = math.atan2(segment_vector_zx[1], segment_vector_zx[0])
        ap_angles.append(ap_angle)
        # projecting the vector in the frontal plane is just removing
        #  the anterioposterior coordinate.
        segment_vector_zy = segment_vector[[2, 1]]
        # for the angle, z is used as the first direction/axis and y as the other
        mp_angle = math.atan2(segment_vector_zy[1], segment_vector_zy[0])
        mp_angles.append(mp_angle)
        distances.append(np.linalg.norm(segment_vector))
    return ap_angles, mp_angles, distances


def get_vectors_joining_points(
    point_layer: "napari.layers.Points",
) -> np.array:
    """Returns an array of the vectors joining each pair of consecutive points.

    Receives a napari points layer whose data attribute must be
    (n_points, n_dims) and returns an array of the vectors joining each pair of
    consecutive points.

    Args:
        point_layer: Napari points layer with at least two points. Shape
            of point_layer.data must be (n_points, n_dims).
    Returns:
        A numpy array of shape (n_points - 1, 2, n_dims) where the first
        dimension corresponds to the 'n_points' - 1 lines that join all
        the pairs of consecutive points in point_layer.data. The second
        dimension, is a 2 because it contains the initial and end point
        of the line.
    """
    points_array = point_layer.data
    n_points = points_array.shape[0]
    if n_points < 2:
        raise ValueError(
            "Expected at least two points in order to run the plugin"
        )

    pointsb = points_array[1:]
    pointsa = points_array[:-1]

    arrays = np.stack([pointsa, pointsb], axis=1)
    return arrays


@magic_factory(
    auto_call=False,
    output_folder={"mode": "d"},
    call_button=True,
    save_resuts_to_output_folder={"label": "Save results to output folder?"},
)
def analyze_points_layer(
    point_layer: "napari.layers.Points",
    save_resuts_to_output_folder: bool = False,
    output_folder: pathlib.Path = "results_folder",
) -> List[napari.types.LayerDataTuple]:
    """Analyzes a point layer and saves distances and angles for consecutive points
    points.csv
    id, coordx,  coordy,  coordz
    0,   ..
    1,   ..
    measurements.csv

    the angles define the vector from point 1 to point2

    point1_id, point2_id, distances, AP_angle, MP_angle
    0,         ..
    1,         ..

    Args:
        point_layer: A points layer to analyze.
        save_resuts_to_output_folder: A boolean, whether to save the results
            to the output folder below. If False, output_folder is ignored.
        output_folder: Name of the folder where the output is going to be saved.
    """
    if save_resuts_to_output_folder and not os.path.exists(output_folder):
        os.mkdir(output_folder)

    assert len(point_layer.data) > 0, "Not enough points in the layer"

    (
        ap_angles,
        mp_angles,
        distances,
    ) = get_angles_and_distances_between_consecutive_points(point_layer)

    points_array = point_layer.data
    n_points = points_array.shape[0]
    n_coords = points_array.shape[1]

    if save_resuts_to_output_folder:
        with open(
            os.path.join(output_folder, "measurements.csv"), "w", newline=""
        ) as csvfile:
            writer = csv.writer(csvfile, delimiter=",")
            writer.writerow(
                ("point1_id", "point2_id", "AP_angle", "MP_angle", "distance")
            )
            for i in range(len(mp_angles)):
                writer.writerow(
                    (i, i + 1, ap_angles[i], mp_angles[i], distances[i])
                )

        with open(
            os.path.join(output_folder, "points.csv"), "w", newline=""
        ) as csvfile:
            writer = csv.writer(csvfile, delimiter=",")
            writer.writerow(
                ("point_id", *(f"coord_{i}" for i in range(n_coords)))
            )
            for i in range(n_points):
                writer.writerow((i, *points_array[i]))

        print(f"created analysis files at folder called: {output_folder}")

    lines_joining_consecutive_points = get_vectors_joining_points(point_layer)

    vectors_props = {
        "distance": distances,
        "ap_angle": ap_angles,
        "mp_angle": mp_angles,
    }

    # Adding vectors as a shapes layer.
    vectors_text = {
        "text": "distance: {distance:.4f}\nap_angle: {ap_angle:.4f}\n mp_angle: {mp_angle:.4f}",
        "size": 8,
        "color": "green",
        "anchor": "upper_left",
        "translation": [-2, 0.0, 0.0]
        if point_layer.data.shape[1] == 3
        else [-2, 0.0],
    }

    lines_layer_meta = {
        "shape_type": "path",
        "edge_color": "blue",
        "properties": vectors_props,
        "text": vectors_text,
        "edge_width": 0.1,
    }

    # Adding a new point layers with the text to show the index of each point.
    points_annotations_properties = {"index": list(range(1, n_points + 1))}

    points_annotations_text = {
        "text": "Point {index}",
        "size": 8,
        "color": "red",
        "anchor": "upper_left",
        "translation": [0.0, 0.0, 0.0]
        if point_layer.data.shape[1] == 3
        else [0.0, 0.0],
    }

    numbered_points_meta = {
        "properties": points_annotations_properties,
        "text": points_annotations_text,
        "edge_color": "transparent",
        "face_color": "transparent",
    }

    return [
        (lines_joining_consecutive_points, lines_layer_meta, "shapes"),
        (point_layer.data, numbered_points_meta, "points"),
    ]



import numpy as np
from scipy.interpolate import splprep, splev


def spline_fit(points, smoothing=0.2, k=3, n_points=100):
    """Given an input set of 2/3D points, returns a new set of points
    representing the spline interpolation
    Parameters
    ----------
    points : np.ndarray
        2/3D array of points defining a path
    smoothing : float
        Smoothing factor
    k : int
        Spline degree
    n_points : int
        How many points used to define the resulting interpolated path
    Returns
    ----------
    new_points : np.ndarray
        Points defining the interpolation
    """

    # scale smoothing to the spread of the points
    max_range = max(np.max(points, axis=0) - np.min(points, axis=0))
    smoothing *= max_range

    # calculate bspline representation
    tck, _ = splprep(points.T, s=smoothing, k=k)

    # evaluate bspline
    spline_fit_points = splev(np.linspace(0, 1, n_points), tck)

    return np.array(spline_fit_points).T



from napari_plugin_engine import napari_hook_implementation
from brainreg_segment.segment import SegmentationWidget
from brainreg_segment.measurement import measurement
from brainreg_segment.measurement.random_layers import (
    generate_random_points_2d_layer,
    generate_random_points_3d_layer,
)


@napari_hook_implementation
def napari_experimental_provide_dock_widget():
    return [
        (SegmentationWidget, {"name": "Manual segmentation"}),
        (measurement.analyze_points_layer, {"name": "Analyze points layer"}),
    ]


@napari_hook_implementation
def napari_provide_sample_data():
    return {
        "random 3d points": generate_random_points_3d_layer,
        "random 2d points": generate_random_points_2d_layer,
    }



import numpy as np
from scipy.spatial import cKDTree


def create_KDTree_from_image(image, value=0):
    """
    Create a KDTree of points equalling a given value
    :param image: Image to be converted to points
    :param value: Value of image to be used
    :return: scipy.spatial.cKDTree object
    """

    list_points = np.argwhere((image == value))
    return cKDTree(list_points)



import imio
import numpy as np

from pathlib import Path
from skimage import measure

from imlib.general.pathlib import append_to_pathlib_stem


def convert_obj_to_br(verts, faces, voxel_size):
    if voxel_size != 1:
        verts = verts * voxel_size

    faces = faces + 1
    return verts, faces


def extract_and_save_object(
    image, output_file_name, voxel_size, threshold=0, step_size=1
):
    verts, faces, normals, values = measure.marching_cubes(
        image, threshold, step_size=step_size
    )
    verts, faces = convert_obj_to_br(verts, faces, voxel_size)
    marching_cubes_to_obj(
        (verts, faces, normals, values), str(output_file_name)
    )


def marching_cubes_to_obj(marching_cubes_out, output_file):
    """
    Saves the output of skimage.measure.marching_cubes as an .obj file
    :param marching_cubes_out: tuple
    :param output_file: str
    """

    verts, faces, normals, _ = marching_cubes_out
    with open(output_file, "w") as f:
        for item in verts:
            f.write(f"v {item[0]} {item[1]} {item[2]}\n")
        for item in normals:
            f.write(f"vn {item[0]} {item[1]} {item[2]}\n")
        for item in faces:
            f.write(
                f"f {item[0]}//{item[0]} {item[1]}//{item[1]} "
                f"{item[2]}//{item[2]}\n"
            )
        f.close()


def volume_to_vector_array_to_obj_file(
    image,
    output_path,
    voxel_size=50,
    step_size=1,
    threshold=0,
    deal_with_regions_separately=False,
):
    if deal_with_regions_separately:
        for label_id in np.unique(image):
            if label_id != 0:
                filename = append_to_pathlib_stem(
                    Path(output_path), "_" + str(label_id)
                )
                image = image == label_id
                extract_and_save_object(
                    image,
                    filename,
                    voxel_size,
                    threshold=threshold,
                    step_size=step_size,
                )
    else:
        extract_and_save_object(
            image,
            output_path,
            voxel_size,
            threshold=threshold,
            step_size=step_size,
        )


def save_label_layers(regions_directory, label_layers):
    print(f"Saving regions to: {regions_directory}")
    regions_directory.mkdir(parents=True, exist_ok=True)
    for label_layer in label_layers:
        save_regions_to_file(label_layer, regions_directory)


def export_label_layers(
    regions_directory, label_layers, voxel_size, obj_ext=".obj"
):
    print(f"Exporting regions to: {regions_directory}")
    regions_directory.mkdir(parents=True, exist_ok=True)
    for label_layer in label_layers:
        filename = regions_directory / (label_layer.name + obj_ext)
        export_regions_to_file(label_layer.data, filename, voxel_size)


def save_regions_to_file(
    label_layer,
    destination_directory,
    ignore_empty=True,
    image_extension=".tiff",
):
    """
    Saves the segmented regions to file (as .tiff)
    :param label_layer: napari labels layer (with segmented regions)
    :param destination_directory: Where to save files to
    :param ignore_empty: If True, don't attempt to save empty images
    :param image_extension: File extension fo the image files
    """
    data = label_layer.data
    if ignore_empty:
        if data.sum() == 0:
            return

    name = label_layer.name

    filename = destination_directory / (name + image_extension)
    imio.to_tiff(data.astype(np.int16), filename)


def export_regions_to_file(image, filename, voxel_size, ignore_empty=True):
    """
    Export regions as .obj for brainrender

    """
    if ignore_empty:
        if image.sum() == 0:
            return

    volume_to_vector_array_to_obj_file(image, filename, voxel_size=voxel_size)



# RegionSeg
from qtpy.QtWidgets import (
    QGridLayout,
    QGroupBox,
)

from brainreg_segment.layout.utils import display_warning
from brainreg_segment.layout.gui_elements import (
    add_button,
    add_checkbox,
)

from brainreg_segment.regions.layers import (
    add_existing_region_segmentation,
    add_new_region_layer,
)

from brainreg_segment.regions.analysis import region_analysis
from brainreg_segment.layout.gui_constants import (
    COLUMN_WIDTH,
    SEGM_METHODS_PANEL_ALIGN,
    CALCULATE_VOLUMES_DEFAULT,
    SUMMARIZE_VOLUMES_DEFAULT,
    BRUSH_SIZE,
    IMAGE_FILE_EXT,
    NUM_COLORS,
)


class RegionSeg(QGroupBox):
    """
    Region segmentation method panel

    """

    def __init__(
        self,
        parent,
        calculate_volumes_default=CALCULATE_VOLUMES_DEFAULT,
        summarise_volumes_default=SUMMARIZE_VOLUMES_DEFAULT,
        brush_size=BRUSH_SIZE,
        image_file_extension=IMAGE_FILE_EXT,
        num_colors=NUM_COLORS,
    ):

        super(RegionSeg, self).__init__()
        self.parent = parent

        self.calculate_volumes_default = calculate_volumes_default
        self.summarise_volumes_default = summarise_volumes_default

        # Brushes / ...
        self.brush_size_default = BRUSH_SIZE  # Keep track of default
        self.brush_size = brush_size  # Initialise
        self.num_colors = num_colors

        # File formats
        self.image_file_extension = image_file_extension

    def add_region_panel(self, row):
        self.region_panel = QGroupBox("Region analysis")
        region_layout = QGridLayout()

        add_button(
            "Add region",
            region_layout,
            self.add_region,
            2,
            0,
        )

        add_button(
            "Analyse regions",
            region_layout,
            self.run_region_analysis,
            2,
            1,
        )

        self.calculate_volumes_checkbox = add_checkbox(
            region_layout,
            self.calculate_volumes_default,
            "Calculate volumes",
            0,
        )

        self.summarise_volumes_checkbox = add_checkbox(
            region_layout,
            self.summarise_volumes_default,
            "Summarise volumes",
            1,
        )

        region_layout.setColumnMinimumWidth(1, COLUMN_WIDTH)
        self.region_panel.setLayout(region_layout)
        self.parent.layout.addWidget(self.region_panel, row, 0, 1, 2)
        self.region_panel.setVisible(False)

    def toggle_region_panel(self):
        # TODO: Change color scheme directly when theme is switched
        # TODO: "text-align" property should follow constant SEGM_METHODS_PANEL_ALIGN
        if self.region_panel.isVisible():
            self.region_panel.setVisible(False)
            if self.parent.viewer.theme == "dark":
                self.parent.show_regionseg_button.setStyleSheet(
                    f"QPushButton {{ background-color: #414851; text-align:{SEGM_METHODS_PANEL_ALIGN};}}"
                    f"QPushButton:pressed {{ background-color: #414851; text-align:{SEGM_METHODS_PANEL_ALIGN};}}"
                )
            else:
                self.parent.show_regionseg_button.setStyleSheet(
                    f"QPushButton {{ background-color: #d6d0ce; text-align:{SEGM_METHODS_PANEL_ALIGN};}}"
                    f"QPushButton:pressed {{ background-color: #d6d0ce; text-align:{SEGM_METHODS_PANEL_ALIGN};}}"
                )

        else:
            self.region_panel.setVisible(True)
            if self.parent.viewer.theme == "dark":
                self.parent.show_regionseg_button.setStyleSheet(
                    f"QPushButton {{ background-color: #7e868f; text-align:{SEGM_METHODS_PANEL_ALIGN};}}"
                    f"QPushButton:pressed {{ background-color: #7e868f; text-align:{SEGM_METHODS_PANEL_ALIGN};}}"
                )
            else:
                self.parent.show_regionseg_button.setStyleSheet(
                    f"QPushButton {{ background-color: #fdf194; text-align:{SEGM_METHODS_PANEL_ALIGN};}}"
                    f"QPushButton:pressed {{ background-color: #fdf194; text-align:{SEGM_METHODS_PANEL_ALIGN};}}"
                )

    def check_saved_region(self):
        add_existing_region_segmentation(
            self.parent.paths.regions_directory,
            self.parent.viewer,
            self.parent.label_layers,
            self.image_file_extension,
        )

    def add_region(self):
        print("Adding a new region\n")
        self.region_panel.setVisible(True)  # Should be visible by default!
        add_new_region_layer(
            self.parent.viewer,
            self.parent.label_layers,
            self.parent.base_layer.data,
            self.brush_size,
            self.num_colors,
        )

    def run_region_analysis(self):
        if self.parent.label_layers:
            choice = display_warning(
                self.parent,
                "About to analyse regions",
                "Existing files will be will be deleted. Proceed?",
            )
            if choice:
                print("Running region analysis")
                worker = region_analysis(
                    self.parent.label_layers,
                    self.parent.atlas_layer.data,
                    self.parent.atlas,
                    self.parent.hemispheres_data,
                    self.parent.paths.regions_directory,
                    output_csv_file=self.parent.paths.region_summary_csv,
                    volumes=self.calculate_volumes_checkbox.isChecked(),
                    summarise=self.summarise_volumes_checkbox.isChecked(),
                )
                worker.start()
            else:
                print("Preventing analysis as user chose 'Cancel'")
        else:
            print("No regions found")



from setuptools import setup, find_namespace_packages
from os import path

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

requirements = [
    "numpy",
    "tables",
    "scikit-image",
    "pandas",
    "napari>=0.4.5",
    "napari-plugin-engine >= 0.1.4",
    "imlib >= 0.0.26",
    "dask >= 2.15.0",
    "imio",
    "brainglobe-napari-io",
]


setup(
    name="brainreg-segment",
    version="0.2.14",
    author="Adam Tyson, Horst Obenhaus",
    author_email="code@adamltyson.com",
    license="BSD-3-Clause",
    description="Manual segmentation of 3D brain structures in a common anatomical space",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=requirements,
    extras_require={
        "dev": [
            "black",
            "pytest-cov",
            "pytest",
            "pytest-qt",
            "coverage",
            "bump2version",
            "pre-commit",
            "flake8",
        ]
    },
    url="https://brainglobe.info/",
    project_urls={
        "Source Code": "https://github.com/brainglobe/brainreg-segment",
        "Bug Tracker": "https://github.com/brainglobe/brainreg-segment/issues",
        "Documentation": "https://docs.brainglobe.info/brainreg-segment",
        "User Support": "https://forum.image.sc/tag/brainglobe",
    },
    python_requires=">=3.7",
    packages=find_namespace_packages(exclude=("docs", "tests*")),
    entry_points={
        "console_scripts": [
            "brainreg-segment = brainreg_segment.segment:main",
        ],
        "napari.plugin": ["brainreg-segment = brainreg_segment.plugins"],
    },
    include_package_data=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Operating System :: POSIX :: Linux",
        "Operating System :: Microsoft :: Windows :: Windows 10",
        "Framework :: napari",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
    ],
    zip_safe=False,
)



# Small layout utils
# import napari
from napari.viewer import Viewer

import bg_space as bg

# from brainreg_segment.layout.gui_constants import ORIENTATIONS

from qtpy.QtWidgets import QMessageBox


def disable_napari_btns(viewer):
    """
    Disable some Napari functions to hide them from user interation
    - Transpose TODO: Understand how to add this properly with space conventions
    - Grid view
    - Console
    - New labels layer
    - New points layer
    - New shapes layer

    """
    viewer.window.qt_viewer.viewerButtons.transposeDimsButton.setVisible(False)
    viewer.window.qt_viewer.viewerButtons.gridViewButton.setVisible(False)
    viewer.window.qt_viewer.viewerButtons.consoleButton.setVisible(False)
    viewer.window.qt_viewer.layerButtons.newLabelsButton.setVisible(False)
    viewer.window.qt_viewer.layerButtons.newPointsButton.setVisible(False)
    viewer.window.qt_viewer.layerButtons.newShapesButton.setVisible(False)


def disable_napari_key_bindings():
    """
    Disable some default key bingings that are unused
    """

    @Viewer.bind_key("Control-G", overwrite=True)
    def no_grid_mode_warning(self):
        print("Grid mode is not supported")

    @Viewer.bind_key("Control-T", overwrite=True)
    def no_tranpose_warning(self):
        print("Transposing is not supported")


def get_dims_from_origins(origins):
    """From a list of BG space abbreviations (e.g. ["asl","sla","lsa"]) get correct axes for display in Napari"""
    all_dims = []
    for o in range(len(origins)):
        sc = bg.AnatomicalSpace(origins[0])
        next_orientation = origins[(o + 1) % len(origins)]
        dims, flips, _, _ = sc.map_to(next_orientation)
        assert not any(
            flips
        ), f"\nReceived orientations: {origins}\nThese require (orientation) flips. This is not currently supported"
        all_dims.append(list(dims))
    return all_dims


def display_warning(widget, title, message):
    """
    Display a warning in a pop up that informs
    about overwriting files
    """
    message_reply = QMessageBox.question(
        widget,
        title,
        message,
        QMessageBox.Yes | QMessageBox.Cancel,
    )
    if message_reply == QMessageBox.Yes:
        return True
    else:
        return False


# def overwrite_napari_roll(viewer):
#     """
#     Overwrite Napari _roll() function with something that makes more sense for a (mouse) brain
#     Goal: Cycle through views (e.g. asl -> lsa -> sal)
#     """
#     dims = get_dims_from_origins(ORIENTATIONS)
#
#     def _roll(self):
#         """Roll order of dimensions for display."""
#         dims_idx = dims.index(self.order)
#         next_dims = dims[(dims_idx + 1) % len(dims)]
#         self.order = next_dims
#
#     # Substitute Napari function
#     napari.components.dims.Dims._roll = _roll



# Constants used throughout
WINDOW_HEIGHT = 750
WINDOW_WIDTH = 1500
COLUMN_WIDTH = 150

DISPLAY_REGION_INFO = (
    True  # Display brain region info string in bottom right corner
)

LOADING_PANEL_ALIGN = (
    "left"  # Alignment of text in pushbuttons in loading panel
)
SEGM_METHODS_PANEL_ALIGN = (
    "left"  # Alignment of text in pushbuttons in methods chooser panel
)

POINT_SIZE = 100
SPLINE_SIZE = 50
NUM_COLORS = 10
BRUSH_SIZE = 250
SPLINE_POINTS_DEFAULT = 1000
SPLINE_SMOOTHING_DEFAULT = 0.1
FIT_DEGREE_DEFAULT = 3

SUMMARISE_TRACK_DEFAULT = True
CALCULATE_VOLUMES_DEFAULT = True
SUMMARIZE_VOLUMES_DEFAULT = True

TRACK_FILE_EXT = ".points"
IMAGE_FILE_EXT = ".tiff"
BOUNDARIES_STRING = "Boundaries"

# ORIENTATIONS = [
#     "asl",
#     "lsa",
#     "sla",
# ]  # When rolling through orientations of the brain, adhere to this order



import tifffile
import numpy as np

from glob import glob
from pathlib import Path


def add_new_label_layer(
    viewer,
    base_image,
    name="region",
    selected_label=1,
    num_colors=10,
    brush_size=30,
):
    """
    Takes an existing napari viewer, and adds a blank label layer
    (same shape as base_image)
    :param viewer: Napari viewer instance
    :param np.array base_image: Underlying image (for the labels to be
    referencing)
    :param str name: Name of the new labels layer
    :param int selected_label: Label ID to be preselected
    :param int num_colors: How many colors (labels)
    :param int brush_size: Default size of the label brush
    :return label_layer: napari labels layer
    """
    labels = np.empty_like(base_image)
    label_layer = viewer.add_labels(labels, num_colors=num_colors, name=name)
    label_layer.n_dimensional = True
    label_layer.selected_label = selected_label
    label_layer.brush_size = brush_size
    return label_layer


def add_new_region_layer(
    viewer, label_layers, image_like, brush_size, num_colors
):
    num = len(label_layers)
    new_label_layer = add_new_label_layer(
        viewer,
        image_like,
        name=f"region_{num}",
        brush_size=brush_size,
        num_colors=num_colors,
    )
    new_label_layer.mode = "PAINT"
    label_layers.append(new_label_layer)


def add_existing_label_layers(
    viewer,
    label_file,
    selected_label=1,
    num_colors=10,
    brush_size=30,
):
    """
    Loads an existing image as a napari labels layer
    :param viewer: Napari viewer instance
    :param label_file: Filename of the image to be loaded
    :param int selected_label: Label ID to be preselected
    :param int num_colors: How many colors (labels)
    :param int brush_size: Default size of the label brush
    :return label_layer: napari labels layer
    """
    label_file = Path(label_file)
    labels = tifffile.imread(label_file)
    label_layer = viewer.add_labels(
        labels, num_colors=num_colors, name=label_file.stem
    )
    label_layer.selected_label = selected_label
    label_layer.brush_size = brush_size
    return label_layer


def add_existing_region_segmentation(
    directory, viewer, label_layers, file_extension
):
    label_files = glob(str(directory) + "/*" + file_extension)
    if directory and label_files != []:
        for label_file in label_files:
            label_layers.append(add_existing_label_layers(viewer, label_file))



import numpy as np
import pandas as pd

from pathlib import Path

from brainreg_segment.tracks import IO

tracks_dir = Path.cwd() / "tests" / "data" / "tracks"

pts_2d = np.array([[53, 748], [68, 745], [77, 747], [87, 748], [97, 749]])


spline = np.array(
    (
        [100.0, 49.81249744, 50.08749819],
        [100.0, 58.06928727, 58.53487443],
        [100.0, 66.38765117, 66.9206766],
        [100.0, 74.74999655, 75.26249729],
        [100.0, 83.13873081, 83.57792909],
        [100.0, 91.53626138, 91.8845646],
        [100.0, 99.92499565, 100.1999964],
        [100.0, 108.28734103, 108.54181709],
        [100.0, 116.60570493, 116.92761926],
        [100.0, 124.86249475, 125.3749955],
    )
)

ATLAS_RESOLUTION = 50


def test_save_single_track_layer(tmpdir, rtol=1e-10):
    points = pd.read_hdf(tracks_dir / "track.points")
    IO.save_single_track(points, "track", tmpdir)

    points_test = pd.read_hdf((tmpdir / "track.points"))
    np.testing.assert_allclose(points, points_test, rtol=rtol)


def test_export_single_spline(tmpdir):
    IO.export_single_spline(spline, "track", tmpdir, ATLAS_RESOLUTION)

    spline_test = np.load(str(tmpdir / "track.npy"))
    spline_validate = np.load(str(tracks_dir / "track.npy"))

    assert (spline_test == spline_validate).all()



class Paths:
    """
    A single class to hold all file paths that may be used.
    """

    def __init__(
        self, brainreg_directory, standard_space=True, atlas_space=False
    ):
        self.brainreg_directory = brainreg_directory
        self.main_directory = self.brainreg_directory / "manual_segmentation"

        if atlas_space:
            self.segmentation_directory = self.main_directory / "atlas_space"
        else:
            if standard_space:
                self.segmentation_directory = (
                    self.main_directory / "standard_space"
                )
            else:
                self.segmentation_directory = (
                    self.main_directory / "sample_space"
                )

        self.regions_directory = self.join_seg_files("regions")
        self.region_summary_csv = self.regions_directory / "summary.csv"

        self.tracks_directory = self.join_seg_files("tracks")

    def join_seg_files(self, filename):
        return self.segmentation_directory / filename



import numpy as np
import pandas as pd


from napari.qt.threading import thread_worker
from skimage.measure import regionprops_table

from imlib.pandas.misc import initialise_df
from imlib.general.list import unique_elements_lists

from brainreg_segment.atlas.utils import lateralise_atlas_image


@thread_worker
def region_analysis(
    label_layers,
    atlas_layer_image,
    atlas,
    hemispheres,
    regions_directory,
    output_csv_file=None,
    volumes=True,
    summarise=True,
):
    regions_directory.mkdir(parents=True, exist_ok=True)
    if volumes:
        print("Calculating region volume distribution")
        print(f"Saving summary volumes to: {regions_directory}")
        for label_layer in label_layers:
            analyse_region_brain_areas(
                label_layer,
                atlas_layer_image,
                hemispheres,
                regions_directory,
                atlas,
            )
    if summarise:
        if output_csv_file is not None:
            print("Summarising regions")
            summarise_brain_regions(
                label_layers, output_csv_file, atlas.resolution
            )

    print("Finished!\n")


def summarise_brain_regions(label_layers, filename, atlas_resolution):
    summaries = []
    for label_layer in label_layers:
        summaries.append(summarise_single_brain_region(label_layer))

    result = pd.concat(summaries)
    # TODO: use atlas.space to make these more intuitive
    volume_header = "volume_mm3"
    length_columns = [
        "axis_0_min_um",
        "axis_1_min_um",
        "axis_2_min_um",
        "axis_0_max_um",
        "axis_1_max_um",
        "axis_2_max_um",
        "axis_0_center_um",
        "axis_1_center_um",
        "axis_2_center_um",
    ]

    result.columns = ["region"] + [volume_header] + length_columns

    voxel_volume_in_mm = np.prod(atlas_resolution) / (1000 ** 3)

    result[volume_header] = result[volume_header] * voxel_volume_in_mm

    for header in length_columns:
        for dim, idx in enumerate(atlas_resolution):
            if header.startswith(f"axis_{dim}"):
                scale = float(idx)
                assert scale > 0
                result[header] = result[header] * scale

    result.to_csv(filename, index=False)


def summarise_single_brain_region(
    label_layer,
    ignore_empty=True,
    properties_to_fetch=[
        "area",
        "bbox",
        "centroid",
    ],
):
    data = label_layer.data
    if ignore_empty:
        if data.sum() == 0:
            return

    regions_table = regionprops_table(
        data.astype(np.uint16), properties=properties_to_fetch
    )
    df = pd.DataFrame.from_dict(regions_table)
    df.insert(0, "Region", label_layer.name)
    return df


def analyse_region_brain_areas(
    label_layer,
    atlas_layer_data,
    hemispheres,
    destination_directory,
    atlas,
    extension=".csv",
    ignore_empty=True,
):
    """

    :param label_layer: napari labels layer (with segmented regions)

    :param ignore_empty: If True, don't analyse empty regions
    """

    data = label_layer.data
    if ignore_empty:
        if data.sum() == 0:
            return

    name = label_layer.name

    masked_annotations = data.astype(bool) * atlas_layer_data

    annotations_left, annotations_right = lateralise_atlas_image(
        masked_annotations,
        hemispheres,
        left_hemisphere_value=atlas.left_hemisphere_value,
        right_hemisphere_value=atlas.right_hemisphere_value,
    )

    unique_vals_left, counts_left = np.unique(
        annotations_left, return_counts=True
    )
    unique_vals_right, counts_right = np.unique(
        annotations_right, return_counts=True
    )
    voxel_volume_in_mm = np.prod(atlas.resolution) / (1000 ** 3)

    df = initialise_df(
        "structure_name",
        "left_volume_mm3",
        "left_percentage_of_total",
        "right_volume_mm3",
        "right_percentage_of_total",
        "total_volume_mm3",
        "percentage_of_total",
    )

    sampled_structures = unique_elements_lists(
        list(unique_vals_left) + list(unique_vals_right)
    )
    total_volume_region = get_total_volume_regions(
        unique_vals_left, unique_vals_right, counts_left, counts_right
    )

    for atlas_value in sampled_structures:
        if atlas_value != 0:
            try:
                df = add_structure_volume_to_df(
                    df,
                    atlas_value,
                    atlas.structures,
                    unique_vals_left,
                    unique_vals_right,
                    counts_left,
                    counts_right,
                    voxel_volume_in_mm,
                    total_volume_voxels=total_volume_region,
                )

            except KeyError:
                print(
                    f"Value: {atlas_value} is not in the atlas structure"
                    f" reference file. Not calculating the volume"
                )
    filename = destination_directory / (name + extension)
    df.to_csv(filename, index=False)


def get_total_volume_regions(
    unique_vals_left,
    unique_vals_right,
    counts_left,
    counts_right,
):
    zero_index_left = np.where(unique_vals_left == 0)[0][0]
    counts_left = list(counts_left)
    counts_left.pop(zero_index_left)

    zero_index_right = np.where(unique_vals_right == 0)[0][0]
    counts_right = list(counts_right)
    counts_right.pop(zero_index_right)

    return sum(counts_left + counts_right)


def add_structure_volume_to_df(
    df,
    atlas_value,
    atlas_structures,
    unique_vals_left,
    unique_vals_right,
    counts_left,
    counts_right,
    voxel_volume,
    total_volume_voxels=None,
):
    name = atlas_structures[atlas_value]["name"]

    left_volume, left_percentage = get_volume_in_hemisphere(
        atlas_value,
        unique_vals_left,
        counts_left,
        total_volume_voxels,
        voxel_volume,
    )
    right_volume, right_percentage = get_volume_in_hemisphere(
        atlas_value,
        unique_vals_right,
        counts_right,
        total_volume_voxels,
        voxel_volume,
    )
    if total_volume_voxels is not None:
        total_percentage = left_percentage + right_percentage
    else:
        total_percentage = 0

    df = df.append(
        {
            "structure_name": name,
            "left_volume_mm3": left_volume,
            "left_percentage_of_total": left_percentage,
            "right_volume_mm3": right_volume,
            "right_percentage_of_total": right_percentage,
            "total_volume_mm3": left_volume + right_volume,
            "percentage_of_total": total_percentage,
        },
        ignore_index=True,
    )
    return df


def get_volume_in_hemisphere(
    atlas_value, unique_vals, counts, total_volume_voxels, voxel_volume
):
    try:
        index = np.where(unique_vals == atlas_value)[0][0]
        volume = counts[index] * voxel_volume
        if total_volume_voxels is not None:
            percentage = 100 * (counts[index] / total_volume_voxels)
        else:
            percentage = 0
    except IndexError:
        volume = 0
        percentage = 0

    return volume, percentage



import numpy as np
import pandas as pd
from pathlib import Path

from bg_atlasapi import BrainGlobeAtlas
from brainreg_segment.tracks.analysis import run_track_analysis


atlas_name = "allen_mouse_50um"
atlas = BrainGlobeAtlas(atlas_name)

tracks_dir = Path.cwd() / "tests" / "data" / "tracks"

spline_validate = np.array(
    (
        [100.0, 49.81249744, 50.08749819],
        [100.0, 58.06928727, 58.53487443],
        [100.0, 66.38765117, 66.9206766],
        [100.0, 74.74999655, 75.26249729],
        [100.0, 83.13873081, 83.57792909],
        [100.0, 91.53626138, 91.8845646],
        [100.0, 99.92499565, 100.1999964],
        [100.0, 108.28734103, 108.54181709],
        [100.0, 116.60570493, 116.92761926],
        [100.0, 124.86249475, 125.3749955],
    )
)


def test_run_track_analysis(tmpdir, rtol=1e-10):
    points = pd.read_hdf(tracks_dir / "track.points")
    spline_test = run_track_analysis(
        points, "track", tmpdir, atlas, spline_points=10
    )

    np.testing.assert_allclose(spline_test, spline_validate, rtol=rtol)

    regions_validate = pd.read_csv(tracks_dir / "track.csv")
    regions_test = pd.read_csv(tmpdir / "track.csv")

    pd.testing.assert_frame_equal(regions_test, regions_validate)
